@@ -28,6 +28,7 @@
 #include <string.h>
 
 #include "gmime-multipart-signed.h"
+#include "gmime-multipart-encrypted.h"
 #include "gmime-message-part.h"
 #include "gmime-stream-filter.h"
 #include "gmime-filter-strip.h"
@@ -37,7 +38,7 @@
 #include "gmime-parser.h"
 #include "gmime-part.h"
 
-#define d(x) x
+#define d(x)
 
 /* GObject class methods */
 static void g_mime_multipart_signed_class_init (GMimeMultipartSignedClass *klass);
@@ -150,7 +151,7 @@ multipart_signed_get_header (GMimeObject *object, const char *header)
 static void
 multipart_signed_remove_header (GMimeObject *object, const char *header)
 {
-	return GMIME_OBJECT_CLASS (parent_class)->remove_header (object, header);
+	GMIME_OBJECT_CLASS (parent_class)->remove_header (object, header);
 }
 
 static void
@@ -222,6 +223,11 @@ sign_prepare (GMimeObject *mime_part)
 	if (GMIME_IS_MULTIPART (mime_part)) {
 		GList *lpart;
 		
+		if (GMIME_IS_MULTIPART_SIGNED (mime_part) || GMIME_IS_MULTIPART_ENCRYPTED (mime_part)) {
+			/* must not modify these parts as they must be treated as opaque */
+			return;
+		}
+		
 		lpart = GMIME_MULTIPART (mime_part)->subparts;
 		while (lpart) {
 			subpart = GMIME_OBJECT (lpart->data);
@@ -341,7 +347,14 @@ g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
 	/* set the content-type of the signature part */
 	content_type = g_mime_content_type_new_from_string (mps->protocol);
 	g_mime_object_set_content_type (signature, content_type);
-	g_mime_part_set_filename (GMIME_PART (signature), "signature.asc");
+	
+	/* FIXME: temporary hack, this info should probably be set in
+	 * the CipherContext class - maybe ::sign can take/output a
+	 * GMimePart instead. */
+	if (!g_ascii_strcasecmp (mps->protocol, "application/pkcs7-signature")) {
+		g_mime_part_set_filename (GMIME_PART (signature), "smime.p7m");
+		g_mime_part_set_encoding (GMIME_PART (signature), GMIME_PART_ENCODING_BASE64);
+	}
 	
 	/* save the content and signature parts */
 	/* FIXME: make sure there aren't any other parts?? */
@@ -372,7 +385,7 @@ g_mime_multipart_signed_sign (GMimeMultipartSigned *mps, GMimeObject *content,
  * fail. If the signing fails, an exception will be set on @err to
  * provide information as to why the failure occured.
  **/
-GMimeCipherValidity *
+GMimeSignatureValidity *
 g_mime_multipart_signed_verify (GMimeMultipartSigned *mps, GMimeCipherContext *ctx,
 				GError **err)
 {
@@ -382,7 +395,7 @@ g_mime_multipart_signed_verify (GMimeMultipartSigned *mps, GMimeCipherContext *c
 	GMimeFilter *crlf_filter;
 	GMimeStream *stream, *sigstream;
 	const char *protocol, *micalg;
-	GMimeCipherValidity *valid;
+	GMimeSignatureValidity *valid;
 	GMimeCipherHash hash;
 	char *content_type;
 	
@@ -390,6 +403,11 @@ g_mime_multipart_signed_verify (GMimeMultipartSigned *mps, GMimeCipherContext *c
 	g_return_val_if_fail (GMIME_IS_CIPHER_CONTEXT (ctx), NULL);
 	g_return_val_if_fail (ctx->sign_protocol != NULL, NULL);
 	
+	if (g_mime_multipart_get_number ((GMimeMultipart *) mps) != 2) {
+		/* FIXME: set an exception */
+		return NULL;
+	}
+	
 	protocol = g_mime_object_get_content_type_parameter (GMIME_OBJECT (mps), "protocol");
 	micalg = g_mime_object_get_content_type_parameter (GMIME_OBJECT (mps), "micalg");
 	
@@ -434,7 +452,18 @@ g_mime_multipart_signed_verify (GMimeMultipartSigned *mps, GMimeCipherContext *c
 	
 	/* get the signature stream */
 	wrapper = g_mime_part_get_content_object (GMIME_PART (signature));
-	sigstream = g_mime_data_wrapper_get_stream (wrapper);
+	
+	/* FIXME: temporary hack for Balsa to support S/MIME,
+	 * ::verify() should probably take a mime part so it can
+	 * decode this itself if it needs to. */
+	if (!g_ascii_strcasecmp (protocol, "application/pkcs7-signature") ||
+	    !g_ascii_strcasecmp (protocol, "application/x-pkcs7-signature")) {
+		sigstream = g_mime_stream_mem_new ();
+		g_mime_data_wrapper_write_to_stream (wrapper, sigstream);
+	} else {
+		sigstream = g_mime_data_wrapper_get_stream (wrapper);
+	}
+	
 	g_mime_stream_reset (sigstream);
 	g_object_unref (signature);
 	g_object_unref (wrapper);
